/*
 * Copyright (c) 2015 LAAS/CNRS
 * All rights reserved.
 *
 * Redistribution and use  in source  and binary  forms,  with or without
 * modification, are permitted provided that the following conditions are
 * met:
 *
 *   1. Redistributions of  source  code must retain the  above copyright
 *      notice and this list of conditions.
 *   2. Redistributions in binary form must reproduce the above copyright
 *      notice and  this list of  conditions in the  documentation and/or
 *      other materials provided with the distribution.
 *
 *                                           Anthony Mallet on Wed Jul  1 2015
 */
#include <sys/time.h>

#include <errno.h>
#include <math.h>
#include <time.h>

#define S_FUNCTION_NAME tk3_rt
#define S_FUNCTION_LEVEL 2

#include "simstruc.h"


/*  --- mdlInitializeSizes ------------------------------------------------- */

/* Abstract:
 *    Determine the S-function block's characteristics:
 *    number of inputs, outputs, states, etc.
 */
static void
mdlInitializeSizes(SimStruct *S)
{
  ssSetNumSFcnParams(S, 1);  /* Number of expected parameters */
  if (ssGetNumSFcnParams(S) != ssGetSFcnParamsCount(S)) {
    return; /* Parameter mismatch reported by the Simulink engine */
  }

  ssSetNumContStates(S, 0);
  ssSetNumDiscStates(S, 1);

  if (!ssSetNumInputPorts(S, 0)) return;
  if (!ssSetNumOutputPorts(S, 2)) return;
  ssSetOutputPortWidth(S, 0, 1);
  ssSetOutputPortWidth(S, 1, 1);

  ssSetNumSampleTimes(S, 1);
  ssSetNumRWork(S, 1);
  ssSetNumIWork(S, 0);
  ssSetNumPWork(S, 1);
  ssSetNumModes(S, 0);
  ssSetNumNonsampledZCs(S, 0);

  ssSetOptions(S, SS_OPTION_EXCEPTION_FREE_CODE);
}

static void
mdlInitializeSampleTimes(SimStruct *S)
{
  ssSetSampleTime(S, 0, CONTINUOUS_SAMPLE_TIME);
  ssSetOffsetTime(S, 0, 0.0);
}

static void
mdlOutputs(SimStruct *S, int_T tid)
{
   double *delay = ssGetOutputPortRealSignal(S, 0);
   double *curtime = ssGetOutputPortRealSignal(S, 1);
   struct timeval *tv0 = ssGetPWorkValue(S, 0);
   struct timespec req, rem;
   struct timeval tv;
   double dt, simdt, tsim, ptsim;
   double catchup;

   UNUSED_ARG(tid); /* not used in single tasking mode */
   if (!tv0) {
     tv0 = malloc(sizeof(*tv0));
     if (tv0) gettimeofday(tv0, NULL);
     ssSetPWorkValue(S, 0, tv0);
     ssSetRWorkValue(S, 0, ssGetT(S));

     delay[0] = 0;
     curtime[0] = tv0->tv_sec + 1e-6*tv0->tv_usec;
     return;
   }


   /* simulation time and real time */
   ptsim = ssGetRWorkValue(S, 0);
   tsim = ssGetT(S);
   ssSetRWorkValue(S, 0, tsim);
   simdt = tsim - ssGetTStart(S);

   gettimeofday(&tv, NULL);
   dt = (tv.tv_sec - tv0->tv_sec) + 1e-6 * (tv.tv_usec - tv0->tv_usec);

   /* sleep time */
   catchup = mxGetScalar(ssGetSFcnParam(S, 0));
   if (catchup < 1. /* skip delays */) {
     while(simdt < dt) dt -= tsim - ptsim;
   }

   /* sleep */
   if (simdt > dt) {
     req.tv_sec = floor(simdt - dt);
     req.tv_nsec = 1e9*(simdt - dt - req.tv_sec);
     while(nanosleep(&req, &rem)) {
       if (errno != EINTR) break;
       req = rem;
     }

     gettimeofday(&tv, NULL);
     dt = (tv.tv_sec - tv0->tv_sec) + 1e-6 * (tv.tv_usec - tv0->tv_usec);
   }

   /* actual delay */
   delay[0] = dt - simdt;
   curtime[0] = tv.tv_sec + 1e-6*tv.tv_usec;
}

static void
mdlTerminate(SimStruct *S)
{
  struct timeval *tv0 = ssGetPWorkValue(S, 0);
  if (tv0) free(tv0);
}

/* Required S-function trailer */
#ifdef  MATLAB_MEX_FILE    /* Is this file being compiled as a MEX-file? */
# include "simulink.c"      /* MEX-file interface mechanism */
#else
# include "cg_sfun.h"       /* Code generation registration function */
#endif
